/*  
file: build_lattice.cpp  
The lattice_graph instance is constructed here.

Spinner v.2, a Monte Carlo program for the simulation of classical
magnetic systems.

Written by Mark de Vries, markadevries@gmail.com

Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of
Edinburgh, Old College, Edinburgh, UK  

Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the 
Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "build_lattice.hpp"

std::ostream& operator<<(std::ostream& os, const atom_data& v)
{
  os<<'\n'<<v.label<<" s = "<<v.s<<"  ("<<v.x[0]<<','<<v.x[1]<<','<<v.x[2]<<")";
  return os;
}

std::vector<atom_data> parse_atoms(const std::vector<std::string>& values)
{
  atom_data entry;
  std::vector<atom_data> unit_cell;
  char_separator<char> sep(";|,)\t ");
  std::vector<std::string>::const_iterator value=values.begin();
  for (;value!=values.end();value++){
    tokenizer<char_separator<char> > atom_tokens(*value,sep);
    tokenizer<char_separator<char> >::iterator it=atom_tokens.begin();
    //    for(it=atom_tokens.begin(); it!=atom_tokens.end();++it) {
      entry.label=*it; it++;
      for (int i=0;i<3;i++) {
        entry.x[i]=atof((*it).c_str()); it++; 
	if (it==atom_tokens.end()) 
	  throw po::invalid_option_value("'add-atom' must contain at least n*4 values, n being the number\nof atoms per unit cell.");
      }
      if (it!=atom_tokens.end()){
	entry.s=atof((*it).c_str()); 
	it++;
      } else entry.s=1;
      if (it!=atom_tokens.end()){
	entry.g=atof((*it).c_str()); 
	it++;
      } else entry.g=2;
      if (it!=atom_tokens.end())
	entry.f=atof((*it).c_str());
      else entry.f=1;
      unit_cell.push_back(entry); 
      //    }
  }
  return unit_cell;
}

std::vector<bond_data> parse_bonds(const std::vector<std::string>& values)
{
  bond_data entry;
  std::vector<bond_data> list_of_bonds;
  char_separator<char> sep(";|,\t ");
  std::vector<std::string>::const_iterator value=values.begin();
  for (;value!=values.end();value++){
    tokenizer<char_separator<char> > atom_tokens(*value,sep);
    tokenizer<char_separator<char> >::iterator it=atom_tokens.begin();
    if (it==atom_tokens.end())
      throw po::invalid_option_value(strcat("'add-bond' must contain 4 entries, exchange strength 'J', bond length, label of site i, label of site j. Found ",(*value).c_str()));    
    entry.J=atof((*it).c_str()); it++;
    if (it==atom_tokens.end())
      throw po::invalid_option_value(strcat("'add-bond' must contain 4 entries, exchange strength 'J', bond length, label of site i, label of site j. Found ",(*value).c_str()));    
    entry.length=atof((*it).c_str());it++;
    if (it==atom_tokens.end())
      throw po::invalid_option_value(strcat("'add-bond' must contain 4 entries, exchange strength 'J', bond length, label of site i, label of site j. Found ",(*value).c_str()));    
    entry.site_i=*it;it++;
    if (it==atom_tokens.end())
      throw po::invalid_option_value(strcat("'add-bond' must contain 4 entries, exchange strength 'J', bond length, label of site i, label of site j. Found ",(*value).c_str()));    
    entry.site_j=*it;
    list_of_bonds.push_back(entry); 
  }
  return list_of_bonds;
}

lattice_graph load_lattice(po::variables_map& vm)
{
  //vm must contain
  //  state-file  - not checked 
  uint na,nb,nc;
  lattice_graph lg;
  std::string state_file,axes,lattice_pars;
  std::fstream fstr_state;
  state_file = vm["state-file"].as<std::string>();

  fstr_state.open(state_file.c_str(),std::ios::in);
  fstr_state>>lg;
  fstr_state>>na>>nb>>nc>>axes>>lattice_pars;
  fstr_state.close();

  std::cout<<"\nLATTICE AND MAGNETIC STRUCTURE LOADED FROM PREVIOUS STATE:\n"<<
    state_file<<"\n\n";
  return lg;
}

lattice_graph build_lattice(po::variables_map& vm)
{
  //vm must contain entries for
  //  load-state     - optional
  //  na,nb,nv       - required if not load-state
  //  periodic-axes  - required if not load-state
  //  lattice-params - required if not load-state
  //  add-atom       - defaults to 0 0 0 1 2 1
  
  //initializing random number generator to create partial occupancy
  //this is achieved simple by setting S=0 with a probability 1-f.       
  //lagged_fibonacci607 fib_gen;
  //uniform_real<double> uni_01_dist(0,1) ;
  //variate_generator<lagged_fibonacci607&, uniform_real<double> > 
  //  uni_01(fib_gen,uni_01_dist);
  //fib_gen.seed(static_cast<uint>(time(0)));

  std::cout.precision(5);
  show_banner_lattice();

  // Check if old state needs to be recovered, otherwise build new lattice.
  if (vm.count("import-state")) return load_lattice(vm);

  // FUNCTION RETURNS HERE IF LATTICE IS LOADED FROM PREVIOUS RUN.

  //construct unit cell (= list of positions and spin of atoms in unit cell)
  //from input parameters.
  std::vector<atom_data> unit_cell;
  if (vm.count("add-atom")) 
    unit_cell=parse_atoms(vm["add-atom"].as<std::vector<std::string> >() );
  else {
    //by default one atom per unit cell, with spin 1.
    std::vector<std::string> default_atoms;
    default_atoms.push_back(".0 .0 .0 1 2 1"); 
    unit_cell=parse_atoms(default_atoms);
  }
    
  //construct lattice vectors from input parameters.
  lattice_vectors lv=get_lattice_vectors(vm["lattice-params"].as<std::string>() );
  std::string bc=vm["periodic-axes"].as<std::string>();
  uint na=vm["La"].as<uint>(); 
  uint nb=vm["Lb"].as<uint>();
  uint nc=vm["Lc"].as<uint>();

  std::cout<<"\nUNIT CELL:"<<unit_cell<<"\n\n"<<"LATTICE VECTORS:\n"<<
    "a "<<lv[A]<<",\tb "<<lv[B]<<",\tc "<<lv[C]<<std::endl;

  //make a graph representing the lattice with a vertex for each atom. 
  uint atom_cnt=unit_cell.size();//atoms per unit cell.
  uint num_sites = na*nb*nc*atom_cnt;
  lattice_graph LG(num_sites);
  std::cout<<"Constructing "<<num_sites<<" atom lattice with "<<
    na<<" x "<<nb<<" x "<<nc<<" unit cells.\n"<<
    "Periodic boundary conditions apply along the axes; "<<
    vm["periodic-axes"].as<std::string>()<<" .\n";

  Site site;
  Edge edge;  
  site_iterator s,t,last_site,first_site;
  edge_iterator edge_iter, last_edge,e,next_edge;
  o_edge_iterator o_edge_it,last_o_edge,oe;
  std::vector<double> distances;

  // In case J1..J4 are used the hamiltonian is normalised so that Jmax=1, in this way
  // we know in which temperature range we can expect out data..
  // This behaviour does not apply when bonds specified using the new add-atom mechanism. 
  double largest_J = std::max(fabs(vm["J1"].as<double>()),fabs(vm["J2"].as<double>()));
  largest_J = std::max(largest_J,fabs(vm["J3"].as<double>()));
  largest_J = std::max(largest_J,fabs(vm["J4"].as<double>()));
  std::cout<<"\nLargest J = "<<largest_J<<"; All magnetic interactions\
 will be \nnormalised relative to this one (unless it is zero).\n";
  if (largest_J == 0) largest_J = 1;
  
  // ------------------------INIT VERTICES------------------------
  // add site coordinates and magnetic properties to graph vertices

  ublas::unit_vector<double> xunit(vm["spin-dimension"].as<int>(),0); 

  // this is an index which is used when fractional occupancies must be 
  // realised.
  std::vector<std::vector<site_iterator> > index_of_sites(unit_cell.size());

  for ( tie(s,last_site)=vertices(LG); s!=last_site; s++){

    //here follows a complicated shuffle with 'div' and 'mod' to calculate 
    //which unit cell (ai,bi,ci) each atom belongs to. 
    uint ai=(( uint(*s)%(na*nb*atom_cnt) )%(na*atom_cnt))/atom_cnt;
    uint bi=(uint(*s)%(na*nb*atom_cnt))/(na*atom_cnt);
    uint ci=uint(*s)/(na*nb*atom_cnt);

    //this is the n'th atom in the unit_cell we are dealing with.
    uint site_index = uint(*s)%atom_cnt;

    //now we can calculate the coordinates of each atom in the unit cell.
    LG[*s].label=unit_cell[site_index].label;
    LG[*s].cell_index[A]=ai; 
    LG[*s].cell_index[B]=bi; 
    LG[*s].cell_index[C]=ci;
    LG[*s].fract_coords = unit_cell[site_index].x;
    LG[*s].x = (LG[*s].cell_index[A]+unit_cell[site_index].x[A]) * lv[A]+
      (LG[*s].cell_index[B]+unit_cell[site_index].x[B]) * lv[B]+
      (LG[*s].cell_index[C]+unit_cell[site_index].x[C]) * lv[C];

    index_of_sites[site_index].push_back(s);

    // |spin| should be 1 by default, and should only be something else
    // if a system with different spin moments is simulated. 
    // mu_eff is calculated as follows:
    // CORRECTION AS OF 9/05/06: 4pi was added in denominator
    // mu_eff=sqrt( mu_0 * (g*mu_b)^2 / (4pi*J*k_B*(1e-10)^3) )
    // mu_0 = 4*pi*1e-7 H/m !!!!CHECK THIS:maybe mu_0 != mu_rest_of_electrons!!!
    // k_B = 1.38066e-23 J/K
    // mu_b = 9.27408e-24 J/T
    // distances will be in units of ngstrm
    // and J in units of Kelvin
    // what i have here does not work out of systems containing magnetic
    // ions with different spin quantum numbers.

    LG[*s].spin=xunit*unit_cell[site_index].s; //DON'T USE S UNEQUAL TO 1 YET.    
    LG[*s].mu_eff = unit_cell[site_index].s * unit_cell[site_index].g * 0.789262 / 
      sqrt(largest_J);
    LG[*s].site_index = site_index;
  }
  
  // If fractional occupancies f(i) are specified for a particular site i, then remove spins for 
  // until a fraction f(i) remains for site i.
  for (int i=0; i<atom_cnt;i++)
    if (unit_cell[i].f < 1) {
      // first shuffle symmetry equivalent sites
      random_shuffle(index_of_sites[i].begin(),index_of_sites[i].end());
      // then remove the last (1-f(i))*100% in the list.
      for (int j=index_of_sites[i].size()-1; j>int(index_of_sites[i].size()*unit_cell[i].f); j--) 
	remove_vertex(*(index_of_sites[i][j]),LG); //remove site from graph
    }
  //there will now be invalid iterators at the end of index_of_sites[i], so this index
  //should not be used below this point.

  if (num_vertices(LG)<num_sites)
    std::cout<<"\nOveral occupancy "<<double(num_vertices(LG))/double(num_sites)*100.<<"%, "<< num_vertices(LG)<<" spins remaining.\nCheck of lattice connectivity supressed."<<std::endl; 
  
  // --------------------------INIT EDGES------------------------
  // add edges to graph and calculate shortest distances between sites 
  // for each edge using the periodic boundary conditions.

  euclid_vector distance,index_difference,translate,cell_offset; 
  bool inserted;

  //read in axis for which periodic boundary conditions (bc) apply
  for ( tie(s,last_site)=vertices(LG); s!=last_site; s++) {
    for (tie(t,last_site)=vertices(LG); t!=last_site; t++) {
      if (!(s==t)) {
	translate=zero; 
	index_difference = LG[*t].cell_index + LG[*t].fract_coords - 
	  LG[*s].cell_index - LG[*s].fract_coords;
	if ( (fabs(index_difference[A]) > (1.0*na)/2.0) & (bc.find("a") < 10 )) 
	  translate[A]=index_difference[A]*na/fabs(index_difference[A]); 
	if ( (fabs(index_difference[B]) > (1.0*nb)/2.0) & (bc.find("b") < 10 )) 
	  translate[B]=index_difference[B]*nb/fabs(index_difference[B]);
	if ( (fabs(index_difference[C]) > (1.0*nc)/2.0) & (bc.find("c") < 10 )) 
	  translate[C]=index_difference[C]*nc/fabs(index_difference[C]);
	
	distance = (LG[*t].x - LG[*s].x) - 
	  (translate[A]*lv[A] + translate[B]*lv[B] + translate[C]*lv[C]);
	cell_offset = LG[*t].cell_index - LG[*s].cell_index - translate;
	
	// Once the length of this edge is found, decide wether to add or not. 
	if (vm.count("cut-off")) { 
	  //only add when distance is > and smaller than cut-off distance.
	  if ((distance[2]>=0)and(ublas::norm_2(distance) < (vm["cut-off"].as<double>()) )) {
	    tie(edge,inserted)=add_edge(*s,*t,LG);
	    if (inserted) {
	      LG[edge].D = zero;
	      LG[edge].ru=distance; 
	      LG[edge].r_abs=fixed(ublas::norm_2(distance),5);
	      LG[edge].ru/=LG[edge].r_abs;
	      LG[edge].cell_offset=cell_offset;
	      if (std::count(distances.begin(),distances.end(),LG[edge].r_abs)==0)
		distances.push_back(LG[edge].r_abs);
	    }
	  }
	}
	else {//add all new edges distance[3]>0.
	  if (distance[2]>=0){
	    tie(edge,inserted)=add_edge(*s,*t,LG);
	    if (inserted){
	      LG[edge].D = zero;
	      LG[edge].ru=distance; 
	      LG[edge].r_abs=fixed(ublas::norm_2(distance),5);
	      LG[edge].ru/=LG[edge].r_abs;
	      LG[edge].cell_offset=cell_offset;
	      if (std::count(distances.begin(),distances.end(),LG[edge].r_abs)==0)
		distances.push_back(LG[edge].r_abs);
	    }
	  }
	}
      }
    }
  }

  //'distances' is a list of all different edge lengths ocurring.
  // we want to sort it so that we can assign J1 to the shortest edges,
  // J2 to the next shortest etc. 
  std::sort(distances.begin(),distances.end());

  // Here J1,J2 .. Jn are assigned. 
  if (vm.count("add-bond")) {
    std::cout<<"\nYou are using the detailed input method for bonds,\n in this \
case bonds are not normalised to the strongest bond.\n\nBONDS:\n";
    std::vector<bond_data> Jn(parse_bonds(vm["add-bond"].as<std::vector<std::string> >()));
    std::vector<bond_data>::iterator J=Jn.begin();

    int n = 0;
    for (; J<Jn.end();J++) {
      ++n;
      std::cout<<"Bond "<<n<<" of strength "<<(*J).J<<", over length "<<(*J).length<<
	", between '"<<(*J).site_i<<"' and '"<<(*J).site_j<<"'.\n";
    }

    for (tie(edge_iter,last_edge)=edges(LG); edge_iter!=last_edge; edge_iter++) {
      LG[*edge_iter].J = 0;
      for (J=Jn.begin(); J<Jn.end();J++)
	if (LG[*edge_iter].r_abs==(*J).length) 
	  if ( ((LG[source(*edge_iter,LG)].label == (*J).site_i) and 
		(LG[target(*edge_iter,LG)].label == (*J).site_j)) or 
	       ((LG[source(*edge_iter,LG)].label == (*J).site_j) and 
		(LG[target(*edge_iter,LG)].label == (*J).site_i)) )
	    {    
	      LG[*edge_iter].J = (*J).J;
	      break;
	    }      
    }
  }
  else 
    for (tie(edge_iter,last_edge)=edges(LG); edge_iter!=last_edge; edge_iter++)
      {
	if (LG[*edge_iter].r_abs==distances[0]) {
	  LG[*edge_iter].J=vm["J1"].as<double>()/largest_J;
	  //LG[*edge_iter].D=zu*vm["Dz"].as<double>();
	}
	else if (LG[*edge_iter].r_abs==distances[1]) 
	  LG[*edge_iter].J=vm["J2"].as<double>()/largest_J;
	else if (LG[*edge_iter].r_abs==distances[2]) 
	  LG[*edge_iter].J=vm["J3"].as<double>()/largest_J;
	else if (LG[*edge_iter].r_abs==distances[3]) 
	  LG[*edge_iter].J=vm["J4"].as<double>()/largest_J;
	else LG[*edge_iter].J = 0;    
      }

  if (vm["Hamiltonian"].as<std::string>().find("dipole-dipole",0) == std::string::npos)
    //Dispose of bonds with J=0, these do nothing but can bog down
    //the simulation speed.
    {
      tie(e,last_edge) = edges(LG);
      for (next_edge=e; e!=last_edge; e = next_edge) {
	++next_edge;
	if (LG[*e].J == 0) remove_edge(*e,LG);
      }
    }
 
  //this block is for DMI.
  std::vector<int> S1site; std::vector<int> S2site;
  std::vector<std::string> cell_offs;
  std::vector<std::string> Dij;
  if (vm.count("DMI")) {
    S1site = vm["S1.site"].as<std::vector<int> >();
    S2site = vm["S2.site"].as<std::vector<int> >();
    cell_offs = vm["cell_offset"].as<std::vector<std::string> >() ;
    Dij = vm["Dij"].as<std::vector<std::string> >();

    if ((S1site.size() != S2site.size())| (S1site.size() != cell_offs.size()) | 
	(S1site.size() != Dij.size()))
      po::invalid_option_value("Dij, S1.site, S2.site, cell_offset must all have the same number of entries."); 
  }

  for (tie(s,last_site)=vertices(LG); s!=last_site; s++) 
    for (tie(o_edge_it,last_o_edge)=out_edges(*s,LG); o_edge_it!=last_o_edge; o_edge_it++)
      for (int n=0; n<S1site.size();n++) //loop over all S1sites specified in config
	if ( LG[*s].site_index==S1site[n] & LG[target(*o_edge_it,LG)].site_index==S2site[n] ) 
	  {
	  if (LG[*o_edge_it].cell_offset == parse_vector(3,cell_offs[n]) )
	    {
	      LG[*o_edge_it].D = parse_vector(3,Dij[n]);
	      LG[*o_edge_it].d = int(source(*o_edge_it,LG));
	    }
	  else if (LG[*o_edge_it].cell_offset == -1* parse_vector(3,cell_offs[n]) )
	    {
	      LG[*o_edge_it].D = parse_vector(3,Dij[n]);
	      LG[*o_edge_it].d = int(source(*o_edge_it,LG));
	    }
	  }

  //-------------------------CHECKUP OF GRAPH-------------------------
  // Here the number of edges are counted for the atoms in the first 
  // unit cell, and all information regarding their connectivity is 
  // displayed. Then the program checks if atoms of a similar symmetry
  // in the other unit cells have the same number of edges.
    
  format conn_table_head
    ("%1% %|5t|%2% %|15t|%3% %|50t|%4% %|57t|%5% %|60t|%6%\n");
  format conn_table
    ("%1% %2% %|5t|(%3%,%4%,%5%)%|15t|(%6%,%7%,%8%)%|50t|%9%%|57t|%10% %|60t|%11%\n");
  std::cout<<"\nNEAREST NEIGHBOURS AT\n"<<distances.size()<<" distance(s) : \n"<<
    distances<<"\n\nAtoms in first unit cell:\n";
  tie(first_site,last_site)=vertices(LG);
  for ( s=first_site; s!=(first_site+unit_cell.size()); s++){
    unit_cell[LG[*s].site_index].number_of_bonds=degree(*s,LG);
    std::cout<<"Number of bonds for this site: "<<degree(*s,LG)<<'\n'<<
      "Out-edges from "<<LG[*s].label<<LG[*s].site_index<<" at "<<
      LG[*s].x<<":\n"<<LG[*s].cell_index<<":\n"<<
      conn_table_head % "to" % "cell"% "ru" % "|dr|" % "J" % "D";
    for (tie(o_edge_it,last_o_edge)=out_edges(*s,LG); o_edge_it!=last_o_edge; 
	 o_edge_it++){
      std::cout<<conn_table
	% LG[target(*o_edge_it,LG)].label % LG[target(*o_edge_it,LG)].site_index
	% fixed(LG[target(*o_edge_it,LG)].cell_index[A],2)  
	% fixed(LG[target(*o_edge_it,LG)].cell_index[B],2) 
	% fixed(LG[target(*o_edge_it,LG)].cell_index[C],2)  
	% LG[*o_edge_it].ru[A] % LG[*o_edge_it].ru[B]
	% LG[*o_edge_it].ru[C] % fixed(LG[*o_edge_it].r_abs,3)         
	% LG[*o_edge_it].J     % LG[*o_edge_it].D;
    }
  }
  
  if (num_sites == num_vertices(LG)) {
    std::cout<<"\nChecking if atoms of same symmetry have same number of"<<
      "\nbonds throughout lattice... ";
    bool ok=true;
    for ( s=first_site+unit_cell.size(); s!=last_site; s++)
      if ( degree(*s,LG) != unit_cell[LG[*s].site_index].number_of_bonds )
	{
	  std::cout<<'\n'<<LG[*s].label<<LG[*s].site_index<<" at "<<
	    LG[*s].cell_index<<" has "<<degree(*s,LG)<<
	    " bonds instead of expected "<<
	    unit_cell[LG[*s].site_index].number_of_bonds<<".\n";
	  ok=false;
	}
    if (!ok) std::cout<<"This might be an indication of an error, but not"<<
	       "if you chose a size > 1 along \nan axis for which periodic"<<
	       " boundary conditions do not apply.\n";
    else std::cout<<"ok\n";
  }
  return LG;
}


