/* 
file : data_handler.hpp
Defines the data_handler class. This class activates the metropolis
algorithm for specified temperatures, and collects and processes
data. 

Spinner is a Monte Carlo program for the simulation of classical
magnetic systems.

Written by Mark de Vries, markadevries@gmail.com

Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of
Edinburgh, Old College, Edinburgh, UK  

Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the 
Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef DATA_HANDLER
#define DATA_HANDLER

#include <iostream>                      // for std::cout
#include <fstream>                 
#include <utility>                       // for std::pair
#include <algorithm>                     // for std::for_each
#include <numeric>
#include <math.h>
#include <vector>
#include <deque>
#include <map>

#include <boost/config.hpp>
#include <boost/utility.hpp>             // for boost::tie
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/format.hpp>

//#define WATCH(a) if (vm.count("verbose")) std::cout << #a" : " << a<<"\t"
//#define REPORT(a) if (vm.count("verbose")) std::cout << a 
//define common header with above defenitions,

#include "math_utility.hpp"
#include "metropolis.hpp"
#include "lattice_graph.hpp"

using namespace boost;
using namespace boost::numeric;
using namespace boost::posix_time;
using namespace boost::gregorian;

class data_handler {
public:
  data_handler(po::variables_map&,map_of_fifos&,lattice_graph&);
  ~data_handler();
  template<class simulator_type,class hamiltonian,class flipper> 
  void run(simulator_type&,hamiltonian&,lattice_graph&,flipper&);
  void save_state(lattice_graph&);
  void load_state(lattice_graph&);
private:
  po::variables_map vm;
  map_of_fifos& dm;
  map_of_fifos::iterator m;
  double dEtot,stdevdE,Sum,Sum2,stability_condition;
  double Eave,E2ave,Mabs,Mabs2,Mabs4;
  double mean, meanofsquare;
  uint sweeps,cnt,probe_interval,probe_points,Nspins; 
  bool average,thermalise;
  std::vector<double>::iterator t;
  std::vector<double> temps;
  std::deque<double> dE;
  std::fstream fstr,fstr_state;
  std::string input_file,output_file,state_file;
  bool energy_stable();
  ptime start_time;
};

data_handler::data_handler(po::variables_map& VM,
			   map_of_fifos& DM,
			   lattice_graph& lg):
  vm(VM),dm(DM),start_time(second_clock::local_time())
{

  Nspins = num_vertices(lg);
  probe_interval = vm["si"].as<uint>();
  probe_points = vm["sc"].as<uint>();
  stability_condition = vm["sf"].as<double>();
  sweeps = vm["sweeps"].as<uint>();
  thermalise = !vm.count("no-thermalise");
  average = !vm.count("no-averaging");

  if (vm.count("T")) 
    temps=parse_doubles(vm["T"].as<std::vector<std::string> >());
  else 
    std::cout<<"No temperatures given; nothing to do here.\n";

  input_file=vm["input-file"].as<std::string>();
  output_file=vm["output-file"].as<std::string>();

  fstr.open(output_file.c_str(), std::ios::out|std::ios::app);
  if (!fstr) throw po::invalid_option_value("output-file cannot be opened.\n");
  std::cout<<"\nDATA OUTPUT :\n"<<output_file;

  if (average) std::cout<<" -averaged data.\n";
  else std::cout<<" -all data, for hystogram analysis.\n";

  if (vm.count("export-state") ) { 
    state_file = vm["state-file"].as<std::string>();
    std::cout<<state_file<<" -will record state of system on exit.\n";
  }

}

template<class simulator_type,class hamiltonian,class flipper>
void data_handler::run(simulator_type& s, 
		       hamiltonian& H,
		       lattice_graph& lg,
		       flipper& flip)
{
  if (temps.size()>0){
    
    format progress_table ("%1% %|8t|%2% %|40t|%3%\n");
    format temp_entry("%|-7g|\t");
    format data_entry("%|=8e|\t");

    std::cout<<"\nMaking "<<sweeps<< " monte carlo steps for each of "<<temps.size()<<
      " temperatures :\n"<<temps<<"\n\n"<<
      "+------------STARTING TIME : "<<start_time<<"-----------+\n";
    
    //thermalisation
    if (thermalise) {
      std::cout<<"Thermalizing system... "<<std::flush;
      do 
	s(100,probe_interval,lg,dm,H,flip); 
      while (!energy_stable());
      std::cout<<"ok\n"<<std::endl;
    }
    //Write headers:
    fstr<<"#start time : "<<start_time;
    fstr<<"\n#spins="<<Nspins<<", #temperatures="<<"\n#T\t";
    //fstr<<temps.size()<<", #J1="<< vm["J1"].as<double>();
    //fstr<<", #J2="<< vm["J2"].as<double>() <<"\n#T\t";
  
    if (average) { 
      std::cout<<"\nDATA HEADER:\n"<<
	"T E Cv |M| |M|^2 |chi| BC {M_i M^2_i chi_i}, i=1..dim\n";
      fstr<<"E\t\tCv\t\t|M|\t\t|M|^2\t\t|chi|\t\tBC\t\tM_x\t\tM^2_x\t\tchi_x\t\tM_y\t\tM^2_y\t\tchi_y\t\tM_z\t\tM^2_z\t\tchi_z\n";
    }
    else { 
      std::cout<<"\nDATA HEADER:\n"<<
	"T E |M| |M|^4 M_i, i=1..dim\n";
      for (m=dm.begin(); m!=dm.end(); m++) 
	fstr<<"<"<< m->first <<">\t";
      fstr<<'\n';
    }
    
    clear(dm.begin(),dm.end());

    std::cout<<"Progress...\n"<<progress_table % "T" %
      "# of equilibr. steps" % "Time elapsed";

    //For all listed temperatures
    for (t=temps.begin(); t!=temps.end(); t++) {
      //Equilibrate
      cnt=0;
      do {
	s(*t,probe_interval,lg,dm,H,flip);
	cnt++;
      }
      while (!energy_stable());

      clear(dm.begin(),dm.end()); 
      //disable the above line to use last data from
      //equilibration as real data. This is valid, since
      //the stability criterium holds for this data already.

      //Do sweeps*N Monte-Carlo steps.
      s(*t,sweeps,lg,dm,H,flip);
     
      //Save averaged data

      if (average) {
        fstr<<temp_entry%(*t);
	Eave = accumulate(dm["E"].begin(),dm["E"].end(),0.0)/dm["E"].size();
	E2ave = inner_product(dm["E"].begin(),dm["E"].end(),dm["E"].begin(),0.0) 
	  / dm["E"].size();
	fstr<<data_entry%(Eave/Nspins);
	fstr<<data_entry%((E2ave-Eave*Eave)/(Nspins*(*t)*(*t)));
	
	Mabs = accumulate(dm["Mabs"].begin(),dm["Mabs"].end(),0.0)/dm["Mabs"].size();
	Mabs2 = inner_product(dm["Mabs"].begin(),dm["Mabs"].end(),
			      dm["Mabs"].begin(),0.0) / dm["Mabs"].size();
	
	Mabs4 = accumulate(dm["Mabs4"].begin(),dm["Mabs4"].end(),0.0)/dm["Mabs4"].size();
			   
	//save data from absolute magnetisation.
	fstr<<data_entry%(Mabs/Nspins);
	fstr<<data_entry%(Mabs2/Nspins);
	fstr<<data_entry%((Mabs2-Mabs*Mabs)/(Nspins*(*t)));
	fstr<<data_entry%(1-Mabs4/(3*Mabs2*Mabs2));//this is the Binder cumulant
			   
	//now handle the remaining entries in dm in a generic manner, skip the
	//first three which were just done.
	m=dm.begin(); m++; m++; m++;
        for (; m!=dm.end(); m++) {
	  mean = accumulate(m->second.begin(),m->second.end(),0.0) / 
	    (m->second.size());
	  meanofsquare = inner_product(m->second.begin(),m->second.end(),
				       m->second.begin(),0.0) 
	    / (m->second.size());
	  fstr<<data_entry%(mean/Nspins);
	  fstr<<data_entry%(meanofsquare/Nspins);
	  fstr<<data_entry%((meanofsquare-mean*mean)/(Nspins*(*t)));
		  
        }
        fstr<<std::endl;
	clear(dm.begin(),dm.end());
      }
      else //Or write away all data for hystogram analysis
        while ( dm.begin()->second.size()>0 ) {
	  fstr<<*t<<'\t';
	  for (m=dm.begin(); m!=dm.end(); m++) {
	    fstr<<m->second[0]<<'\t';
	    m->second.pop_front();
	  }
	  fstr<<'\n';
        }

      //save magnetic structure:

      ptime now = second_clock::local_time();
      time_duration elapsed = now - start_time;
      std::cout<<progress_table % *t % (cnt*probe_interval) % elapsed <<
	std::flush;
    }
    fstr<<"\n\n";
    std::cout<<"\nData is saved to : "<<output_file;
   
  }
  save_state(lg);
}

void data_handler::save_state(lattice_graph& lg)
{
  if (vm.count("export-state")) {
    fstr_state.open(state_file.c_str(),std::ios::out);
    fstr_state<<lg<<"\n"<<
      vm["La"].as<uint>()<<' '<<vm["Lb"].as<uint>()<<' '<<vm["Lc"].as<uint>()<<"\n"<<
      vm["periodic-axes"].as<std::string>()<<"\n"<<
      vm["lattice-params"].as<std::string>()<<"\n";
    fstr_state.close();
    std::cout<<"\nState of system saved to : "<<state_file;
  }
}

bool data_handler::energy_stable()
{
  //Need at least two energy points to compare, at a distance given
  //by 'probe_interval', which is a variable, default = 1.
  if (dm["E"].size() < probe_interval+1 ) return false;

  //Store the energy difference attained the probe_inverval.
  dE.push_back(*(dm["E"].end()-1)-*(dm["E"].end()-1-probe_interval) );

  //Accumulate enough data (probe_points) to test stability.
  if (dE.size() < probe_points) return false;

  //Calculate total change in energy, and standard deviation
  dEtot=accumulate(dE.begin(),dE.end(),0.0);
  stdevdE=sqrt(inner_product(dE.begin(),dE.end(),dE.begin(),0.0)/dE.size());

  if (stability_condition*fabs(dEtot) <= stdevdE)
    { 
      //nominal change in energy is now much smaller than the random 
      //variations => stabilized!
      dE=empty_fifo;
      return true;
    }
  else 
    { 
      //Discard of data at beginning of fifos to make place for the
      //data from the next cycle. 
      for (int i=0; i < probe_interval;i++ ) 
	for (m=dm.begin(); m!=dm.end(); m++) m->second.pop_front();
      dE.pop_front();      
      return false;
    }
}

data_handler::~data_handler() {
    if (fstr) fstr.close(); 
    ptime end_time = second_clock::local_time();
    std::cout<<
      "\n\n+-------------END TIME : "<<end_time<<"---------------+\n";

  }

#endif
