/* 
file: hamiltonians.hpp
Defines the Hamiltonian functors (i.e. classes callable as functions). 

Spinner v.2, a Monte Carlo program for the simulation of 'classical'
magnetic systems.

Written by Mark de Vries, markadevries@gmail.com

Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of
Edinburgh, Old College, Edinburgh, UK  

Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the 
Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/

#ifndef HAMILTONIANS
#define HAMILTONIANS

#include "math_utility.hpp"
#include "options.hpp"

#include <boost/random/lagged_fibonacci.hpp> 
#include <boost/random/variate_generator.hpp>
#include <boost/random/uniform_real.hpp>


struct MolecularField
{
public: 
  MolecularField (po::variables_map& vm) : 
    Hext(3), Hmol(3), uni_01_dist(0,1), 
    uni_01(fib_gen,uni_01_dist)

  {
    Hext = parse_vector(3,vm["Field"].as<std::string>());
    std::cout<<"\nHAMILTONIAN: molecular field, Zeeman term\n";
    std::cout<<"\tField: "<<Hext<<"\n";
    fib_gen.seed(static_cast<uint>(time(0))+1000);
  }
  template<class iterator_type>
  double operator()(iterator_type edge,iterator_type edges_end,
		    site_iterator site,site_iterator sites_end,
		    lattice_graph& LG);
protected:
  euclid_vector Hext,Hmol;
  double E,Habs,phi;
  o_edge_iterator oe;
  lagged_fibonacci607 fib_gen;  
  uniform_real<long double> uni_01_dist;
  variate_generator<lagged_fibonacci607&, uniform_real<long double> > uni_01;

};

template<class iterator_type>
double MolecularField::operator()(iterator_type edge,iterator_type edges_end,
				  site_iterator site,site_iterator sites_end,
				  lattice_graph& LG) 
{ 
  // edge and edges_end are not used here. The iteration is over sites primarily.    
  // for the relationship between the classical and quantum spin see
  // "Modern Quantum Mechanics" by J.J.Sakurai, page 168.
  // Here the situation is much simpler, since we only need to determine the
  // component of the spin paralel and antiparalel to the quantisation axis 
  // defined by "Hmol". This is done using the cosine rule.
  
  E = 0;
  for (;site!=sites_end; site++) {
    Hmol = Hext; //this is the external field.
    
    //Now the molecular field is calculated, summing over classical spins:
    for (oe = out_edges(*site,LG).first; oe!=out_edges(*site,LG).second; oe++)
      Hmol += LG[*oe].J * LG[target(*oe,LG)].spin; 
    Habs = ublas::norm_2(Hmol);
    
    if (Habs != 0) {
      phi = acos((1+pow(Habs,2)-pow(ublas::norm_2(Hmol-LG[*site].spin),2))/(2*Habs));
      E += (uni_01() < pow(cos(phi/2),2)) ? Habs:-Habs;
      //E += Habs*cos(phi); //this is the classical version.
      //if (isnan(E))
      //std::cout<<"phi "<<phi<<" Habs "<<Habs<<" Hmol "<<Hmol<<" E "<<E<<std::endl;
    }
  }
  return E;
}




struct super_exchange 
{
public: 
  super_exchange (po::variables_map& vm) : 
    H(vm["spin-dimension"].as<int>()),
    MAE(vm["spin-dimension"].as<int>()) 
  {
    H = parse_vector(vm["spin-dimension"].as<int>(),vm["Field"].as<std::string>());
    MAE = parse_vector(vm["spin-dimension"].as<int>(),vm["MAE"].as<std::string>());
    std::cout<<"\nHAMILTONIAN: super-exchange, Field, MAE\n";
    std::cout<<"\tField: "<<H<<"\n\tMAE: "<<MAE<<"\n";
  }
  template<class iterator_type>
  double operator()(iterator_type edge,iterator_type edges_end,
		     site_iterator site,site_iterator sites_end,
		     lattice_graph& LG) 
  {
    E = 0;
    for (;edge!=edges_end; edge++)
      E += LG[*edge].J * ublas::inner_prod(LG[source(*edge,LG)].spin,
					   LG[target(*edge,LG)].spin);
    //std::cout<<"SE E = "<<E<<" MAE = "<<MAE<<" spin^2 = "<<ublas::element_prod(LG[*site].spin,LG[*site].spin)<<std::endl;
    for (;site!=sites_end; site++) {
      E += ublas::inner_prod(ublas::element_prod(LG[*site].spin,LG[*site].spin),MAE)+
	ublas::inner_prod(LG[*site].mu_eff*LG[*site].spin,H);
      //std::cout<<" MAE E = "<<E<<std::endl;
    }
    return E;
  }
  euclid_vector MAE,H;
  double E;
};

struct H_Potts
{
public: 
  H_Potts (po::variables_map& vm) : 
    H(vm["spin-dimension"].as<int>()),
    MAE(vm["spin-dimension"].as<int>()) 
  {
    H = parse_vector(vm["spin-dimension"].as<int>(),vm["Field"].as<std::string>());
    MAE = parse_vector(vm["spin-dimension"].as<int>(),vm["MAE"].as<std::string>());
    std::cout<<"\nHAMILTONIAN: Potts, Field, MAE\n";
    std::cout<<"\tField: "<<H<<"\n\tMAE: "<<MAE<<"\n";
  }
  template<class iterator_type>
  double operator()(iterator_type edge,iterator_type edges_end,
		     site_iterator site,site_iterator sites_end,
		     lattice_graph& LG) 
  {
    E = 0;
    for (;edge!=edges_end; edge++)
      E += (LG[source(*edge,LG)].spin == LG[target(*edge,LG)].spin)? LG[*edge].J:0;
    //std::cout<<"SE E = "<<E<<" MAE = "<<MAE<<" spin^2 = "<<ublas::element_prod(LG[*site].spin,LG[*site].spin)<<std::endl;
    for (;site!=sites_end; site++) {
      E += ublas::inner_prod(ublas::element_prod(LG[*site].spin,LG[*site].spin),MAE)+
	ublas::inner_prod(LG[*site].mu_eff*LG[*site].spin,H);
      //std::cout<<" MAE E = "<<E<<std::endl;
    }
    return E;
  }
  euclid_vector MAE,H;
  double E;
};

struct dipole_truncated 
{
public: 
  dipole_truncated (po::variables_map& vm) {	
    std::cout<<"\nHAMILTONIAN: super-exchange, truncated dipole-dipole.\n"; }
  template<class iterator_type>
  double operator()(iterator_type edge,iterator_type edges_end,
		     site_iterator site,site_iterator sites_end,
		     lattice_graph& LG) 
  {
    E = 0; 
    for (;edge!=edges_end; edge++){
      SiSj = ublas::inner_prod(LG[source(*edge,LG)].spin,
                               LG[target(*edge,LG)].spin);
      E += LG[*edge].J * SiSj + pow(LG[*edge].r_abs,-3) *
        LG[source(*edge,LG)].mu_eff * LG[target(*edge,LG)].mu_eff * 
        (SiSj - 3 * ublas::inner_prod(LG[source(*edge,LG)].spin,LG[*edge].ru)* 
                    ublas::inner_prod(LG[target(*edge,LG)].spin,LG[*edge].ru) );
    }
    return E;
  }
  double E, SiSj;
};

struct dipole_truncated_field
{
public: 
  dipole_truncated_field (po::variables_map& vm) : 
    H(parse_vector(3,vm["Field"].as<std::string>() )),
    MAE(parse_vector(3,vm["MAE"].as<std::string>() )) 
  {	
    std::cout<<"\nHAMILTONIAN: super-exchange, truncated dipole-dipole, MAE, field\n"; 
    std::cout<<"\t Field: "<<H<<"\n\tMAE: "<<MAE<<"\n";
  }
  template<class edge_it_type>
  double operator()(edge_it_type edge, edge_it_type edges_end,
		     site_iterator site, site_iterator sites_end,
		     lattice_graph& LG) 
  {
    E = 0; 
    for (;edge!=edges_end; edge++)
      {
	SiSj = ublas::inner_prod(LG[source(*edge,LG)].spin,
				 LG[target(*edge,LG)].spin);
	E += LG[*edge].J * SiSj + pow(LG[*edge].r_abs,-3) *
	  LG[source(*edge,LG)].mu_eff * LG[target(*edge,LG)].mu_eff * 
	  (SiSj - 3 * ublas::inner_prod(LG[source(*edge,LG)].spin,LG[*edge].ru)* 
	   ublas::inner_prod(LG[target(*edge,LG)].spin,LG[*edge].ru) );
      }
    for (;site!=sites_end; site++)
      E += ublas::inner_prod(ublas::element_prod(LG[*site].spin,LG[*site].spin),MAE)+
	ublas::inner_prod(LG[*site].mu_eff*LG[*site].spin,H);
    return E;
  }
  euclid_vector H,MAE;
  double E, SiSj;
};

struct dipole_reaction_field
{
public: 
  dipole_reaction_field (po::variables_map& vm)  
  {	
    std::cout<<"\nHAMILTONIAN: super-exchange, dipole-dipole with reaction field\n"; 
    if (vm.count("cut-off"))
      Rc=vm["cut-off"].as<double>(); 
    else 
      throw po::invalid_option_value("a cut-off radius must be specified in combination with the reaction field Hamiltonian.");
  }
  template<class edge_it_type>
  double operator()(edge_it_type edge, edge_it_type edges_end,
		     site_iterator site, site_iterator sites_end,
		     lattice_graph& LG) 
  {
    E = 0; 
    for (;site!=sites_end; site++) {
      Hr = LG[*site].spin*LG[*site].mu_eff;
      for (o_edge=out_edges(*site,LG).first;o_edge!=out_edges(*site,LG).second;o_edge++) {
	SiSj = ublas::inner_prod(LG[source(*o_edge,LG)].spin,
				 LG[target(*o_edge,LG)].spin);
	E += 0.5*LG[*o_edge].J * SiSj + 0.5*pow(LG[*o_edge].r_abs,-3) *
	  LG[source(*o_edge,LG)].mu_eff * LG[target(*o_edge,LG)].mu_eff * 
	  (SiSj - 3 * ublas::inner_prod(LG[source(*o_edge,LG)].spin,LG[*o_edge].ru)* 
	   ublas::inner_prod(LG[target(*o_edge,LG)].spin,LG[*o_edge].ru) );
	// for each spin calculate reaction field, with epsilon-> infinity.
	Hr += LG[target(*o_edge,LG)].spin*LG[target(*o_edge,LG)].mu_eff;
      }
      //substract the energy of the spin due to the reaction field. 
      E -= 0.5*ublas::inner_prod(LG[*site].mu_eff*LG[*site].spin,Hr/pow(Rc,3));
    }
    return E;
  }
private:
  euclid_vector Hr;
  o_edge_iterator o_edge;
  double E,SiSj,Rc;
};
  

struct dipole_Ewald_sum
{ //TODO: make kappa and kmax vectors, and make pow(K,2) an innerproduct.
  //then kappa and kmax can be added to options. 
public: 
  dipole_Ewald_sum (po::variables_map& vm) { 
    //get kappa,kmax and L from variables map here.
    // and calc. dk.
    std::cout<<"\nHAMILTONIAN: super-exchange, dipole-dipole with Ewald sum.\n"; 
    lv = get_lattice_vectors(vm["lattice-params"].as<std::string>());
    L[A] = vm["La"].as<uint>();     dk[A] = 1/(L[A]*ublas::inner_prod(lv[A],xu)); 
    L[B] = vm["Lb"].as<uint>();     dk[B] = 1/(L[B]*ublas::inner_prod(lv[B],yu)); 
    L[C] = vm["Lc"].as<uint>();     dk[C] = 1/(L[C]*ublas::inner_prod(lv[C],zu));
    //Sampling in k-space goes over a rectangular volume, because this is
    //what the relative distances between sites correspond do, since we use periodic
    //boundary conditions. 
    V = ublas::inner_prod( outer_product( L[A]*lv[A], L[B]*lv[B]),
			   L[C]*lv[C] ); 
    euclid_vector kpoints = parse_vector(3,vm["kpoints"].as<std::string>());
    kmax[A] = dk[A]*kpoints[A]; 
    kmax[B] = dk[B]*kpoints[B];
    kmax[C] = dk[C]*kpoints[C];
    kappa = vm["kappa"].as<double>(); 
    sqrt_pi = sqrt(M_PI);
  }
  template<class iterator_type>
  double operator()(iterator_type edge,iterator_type edges_end,
		     site_iterator site,site_iterator sites_end,
		     lattice_graph& LG) 
  {
    E = 0;
    for (;edge!=edges_end; edge++){
      Si = LG[source(*edge,LG)].spin; //this cannot be out of loop (hint: total energy)
      Sj = LG[target(*edge,LG)].spin;
      r_ij = LG[*edge].r_abs;
      ru_ij = LG[*edge].ru;
      SiSj = ublas::inner_prod(Si,Sj);
      E += LG[*edge].J * SiSj
	+ LG[source(*edge,LG)].mu_eff * LG[target(*edge,LG)].mu_eff
	* ( SiSj * ( erfc(kappa*r_ij) / pow(r_ij,3) + 
		     2*kappa/sqrt_pi*exp(-pow(kappa*r_ij,2))/pow(r_ij,2) + 
		     4*M_PI/(3*V) ) /*surface term*/
	    - ublas::inner_prod(Si,ru_ij) * ublas::inner_prod(Sj,ru_ij) *
	    (3*erfc(kappa*r_ij)/pow(r_ij,3)+2*kappa/sqrt_pi*(2*pow(kappa,2)+3/pow(r_ij,2)) *
	     exp(-pow(kappa*r_ij,2))) + 4 * M_PI / V * Sum_dk() );//sum over K in subroutine
    }
    //self term:
    for (;site!=sites_end; site++)
      E -= 2*pow(kappa,3)*sqr(LG[*site].mu_eff*LG[*site].spin)/(3*sqrt_pi);
    return E;
  }
private:
  double E,SiSj,r_ij,Sum,V,kappa,sqrt_pi;
  euclid_vector Si,Sj,ru_ij,L,nk,K,dk,kmax;
  lattice_vectors lv;
  
  inline double Sum_dk(){
    Sum = 0;
    for (nk[C] = dk[C]; nk[C] < kmax[C]; nk[C] += dk[C])
      for (nk[B] = dk[B]; nk[B] < kmax[B]; nk[B] += dk[B])
	for (nk[A] = dk[A]; nk[A] < kmax[A]; nk[A] += dk[A]){
	  K = nk[A]*xu+nk[B]*yu+nk[C]*zu;
	  Sum += ublas::inner_prod(Si,K)*ublas::inner_prod(Sj,K)/sqr(K)*
	    exp(-sqr(K/(2*kappa))*cos(ublas::inner_prod(K,r_ij*ru_ij)));
	}
    return Sum;
  }
  
};

struct local_MAE
{
public: 
  local_MAE (po::variables_map& vm) :
    //vm(vm),
    a(parse_doubles(vm["a"].as<std::vector<std::string> >() )),
    H(parse_vector(3,vm["Field"].as<std::string>() ))
  {
    euclid_matrix M=M_zero;
    euclid_matrix R,Rt,Mprime;
    std::cout<<"\nHAMILTONIAN: super-exchange, field, local MAE\n"; 
    //read in the euler angles which specify the direction of the anistropy
    //for each site in the unit cell.
    std::vector<double> phi=parse_doubles(vm["phi"].as< std::vector<std::string> >() );
    std::vector<double> theta=parse_doubles(vm["theta"].as< std::vector<std::string> >() );
    std::vector<double> psi=parse_doubles(vm["psi"].as< std::vector<std::string> >() );
    std::vector<double> A_z = parse_doubles(vm["A_z"].as< std::vector<std::string> >() );
    std::vector<double> A_xy = parse_doubles(vm["A_xy"].as<std::vector<std::string> >() );

    std::cout<<"Local anisotropy configuration:\n";
    for (int i=0;i<phi.size();i++){
      int j; if (A_z.size()==1) j=0; else j=i;
      M(0,0) = A_xy[j]; M(1,1) = -A_xy[j]; M(2,2) = A_z[j];
      R=RzRxRz(theta[i], phi[i], psi[i]);
      Rt=ublas::trans(R);//is transpose
      std::cout<<"Ion nr:"<<i<<'\n';
      std::cout<<"R = "<<R<<"\nRt = "<<Rt<<"\nRtMAER = ";
      Mprime = ublas::prod(M,R);
      Ri.push_back(R);
      RtMAER.push_back(ublas::prod(Rt,Mprime));
      euclid_vector S;
      S(0)=1; S(1)=2; S(2)=3;
      std::cout<<RtMAER[i]<<'\n';
    }  
  }
  template<class edge_it_type>
  double operator()(edge_it_type edge, edge_it_type edges_end,
		     site_iterator site, site_iterator sites_end,
		     lattice_graph& LG) 
  {
    E = 0; 
    for (;edge!=edges_end; edge++)
      E += LG[*edge].J * ublas::inner_prod(LG[source(*edge,LG)].spin,
					   LG[target(*edge,LG)].spin);
    for (;site!=sites_end; site++){
      RS = ublas::prod(Ri[LG[*site].site_index],LG[*site].spin);
      E +=
	ublas::inner_prod(LG[*site].spin,ublas::prod(RtMAER[LG[*site].site_index],LG[*site].spin))+
	a[LG[*site].site_index]*(pow(RS[0],4)+pow(RS[1],4)+pow(RS[2],4)) + 
	ublas::inner_prod(LG[*site].mu_eff*LG[*site].spin,H);
    }
    return E;
  }
  //po::variables_map vm;
  euclid_vector H,RS;
  std::vector<double> a;
  std::vector<euclid_matrix> RtMAER,Ri;
  double E, SiSj;
};


struct DMI 
{
public: 
   DMI (po::variables_map& vm) {	
     std::cout<<"\nHAMILTONIAN: super-exchange, DMI interaction.\n";}
   DMI (){  
     std::cout<<"\nHAMILTONIAN: super-exchange, DMI interaction.\n";}	
   template<class iterator_type>
   double operator()(iterator_type edge,iterator_type edges_end,
 		     site_iterator site,site_iterator sites_end,
 		     lattice_graph& LG) 
   {
     E = 0; 
     for (;edge!=edges_end; edge++){
       EJ = LG[*edge].J *ublas::inner_prod(LG[source(*edge,LG)].spin,
					   LG[target(*edge,LG)].spin);
       ED = ublas::inner_prod(LG[*edge].D,outer_product(LG[source(*edge,LG)].spin,
						    LG[target(*edge,LG)].spin));
       //The outer product is antisymmetric on permutation of the spins,
       //so if source and target are reversed, then sign of ED must be reversed. 
       if (int(source(*edge,LG))==LG[*edge].d)        
	 E+=EJ+ED;
       else        
	 E+=EJ-ED;
     }
     return E;
   }
   double E, SiSj,EJ,ED;
 };

struct DMI_Dz 
{
public: 
   DMI_Dz (po::variables_map& vm) {	
     std::cout<<"\nHAMILTONIAN: super-exchange, DMI interaction.\n";}
   DMI_Dz (){  
     std::cout<<"\nHAMILTONIAN: super-exchange, DMI interaction.\n";}	
   template<class iterator_type>
   double operator()(iterator_type edge,iterator_type edges_end,
 		     site_iterator site,site_iterator sites_end,
 		     lattice_graph& LG) 
   {
     E = 0; 
     for (;edge!=edges_end; edge++){
       Si=LG[source(*edge,LG)].spin; 
       Sj=LG[target(*edge,LG)].spin; 
       EJ = LG[*edge].J *ublas::inner_prod(Si,Sj);
       ED = LG[*edge].D[3]*(Si[0]*Sj[1]-Si[1]*Sj[0]);
       E+=EJ+ED;
       /*std::cout<<"EJ="<<EJ<<", ED="<<ED<<", Si="<< LG[source(*edge,LG)].spin<<
	 ", Sj="<<LG[target(*edge,LG)].spin<<
	 " SixSj="<<outer_product(LG[source(*edge,LG)].spin,LG[target(*edge,LG)].spin)<<std::endl;
       */
     }
     return E;
   }
  double E, SiSj,EJ,ED;
  euclid_vector Si,Sj;
  
 };
/*
struct Hz_spin_half
{
public: 
  Hz_spin_half(po::variables_map& vm) : 
    H(parse_vector(3,vm["Field"].as<std::string>() )),
  {
    std::cout<<"\nHAMILTONIAN: Hz and Zeeman term with S=1/2\n";
    std::cout<<"\t Field: "<<H<<"\n";
  }
  template<class iterator_type>
  double& operator()(iterator_type edge,iterator_type edges_end,
		     site_iterator site,site_iterator sites_end,
		     lattice_graph& LG) 
  {
    E = 0;
    for (;edge!=edges_end; edge++)
      E += LG[*edge].J * ublas::inner_prod(LG[source(*edge,LG)].spin,
					   LG[target(*edge,LG)].spin);
    for (;site!=sites_end; site++)
      E += ublas::inner_prod(LG[*site].mu_eff*LG[*site].spin,H);
    return E;
  }
private:
  euclid_vector H;
  double E;
  inline 
};
*/
#endif
