/*
Written by Mark de Vries, markadevries@gmail.com

Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of
Edinburgh, Old College, Edinburgh, UK  

Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the 
Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK

*/
#include "lattice_graph.hpp"

std::ostream& operator<<(std::ostream& ostr,lattice_graph& lg)
{
  //extend this one with output of edges, that is easily done.
  //first number of vertices, then edges; pairs of vertices, then vertex 
  //content, then edge content.
  site_iterator s,last_site;
  std::pair<int,int> edge;
  edge_iterator es,ef;
  format site_table("%1%\t%2%\t%3%\t%4%\t%5%\t%6%\t%7%\t%8%\n");
  format edge_table("%1%\t<->\t%2%\t%3%\t%4%\t%5%\t%6%\t%7%\n");
  
  ostr<<num_vertices(lg)<<'\n';
  for ( tie(s,last_site)=vertices(lg); s!=(last_site); s++)
    ostr<<site_table%int(*s)%lg[*s].label%lg[*s].site_index%lg[*s].cell_index%lg[*s].fract_coords%lg[*s].x%
      lg[*s].mu_eff%lg[*s].spin;

  ostr<<num_edges(lg)<<'\n';
  for ( tie(es,ef)=edges(lg); es!=ef; es++)
    ostr<<edge_table%source(*es,lg)%target(*es,lg)%lg[*es].J%
      lg[*es].r_abs%lg[*es].ru%lg[*es].D%lg[*es].d;
  
  return ostr;
}

lattice_graph& operator>>(std::istream& istr,lattice_graph& lg)
{ //now only compatible with structure_out
  vertex_data site;
  edge_data bond;
  uint source,target,i;
  uint Nspins,Nbonds;

  istr>>Nspins;
  for (uint n=0; n<Nspins; n++){
    istr>>i>>site.label>>site.site_index>>site.cell_index>>site.fract_coords>>site.x>>
      site.mu_eff>>site.spin;
    add_vertex(site,lg);
  }
  istr>>Nbonds;
  std::string c;
  for (uint n=0; n<Nbonds; n++){
    istr>>source>>c>>target>>bond.J>>bond.r_abs>>bond.ru>>bond.D>>bond.d;
    add_edge(source,target,bond,lg);
  }
  return lg;
}

std::string out_VECT(lattice_graph& lg)
{ //to make a file that can be read by the program geomview,
  //for a graphical representation.
  std::ostringstream ostr;
  uint N_spins=num_vertices(lg);
  uint n;
  double startx,starty,startz,dx,dy,dz,veclen=0.4;
  std::string vertices_per_line="";
  std::string colour_changes_per_line="";
  site_iterator s,last_site;

  ostr<<"VECT\n"<<N_spins<<" "<<3*N_spins<<" "<<3*N_spins<<"\n";
  for (n=0; n<N_spins; n++) { 
    vertices_per_line += "3 "; 
    colour_changes_per_line += "3 "; 
  }
  ostr<<vertices_per_line<<" \n"<<colour_changes_per_line<<" \n";

  for ( tie(s,last_site)=vertices(lg); s!=(last_site); s++){
    startx = lg[*s].x[X];
    starty = lg[*s].x[Y];
    startz = lg[*s].x[Z];
    dx = veclen*lg[*s].spin[X];
    dy=0; dz=0;
    if (lg[*s].spin.size()>1) dy = veclen*lg[*s].spin[Y];
    if (lg[*s].spin.size()>2) dz = veclen*lg[*s].spin[Z];
    ostr<<startx<<" "<<starty<<" "<<startz<<"\t"<<startx+dx/2<<" "<<
      starty+dy/2<<" "<<startz+dz/2<<"\t"<<startx+dx<<" "<<
      starty+dy<<" "<<startz+dz<<"\n";
  }
  colour_changes_per_line = "";
  for (n=0; n<N_spins; n++) 
    colour_changes_per_line += "1.0 0.0 0.0 0.0   1.0 1.0 1.0 0.0   1.0 0.0 0.0 0.0\n";
  ostr << colour_changes_per_line;
  return ostr.str();
}

lattice_vectors get_lattice_vectors(const std::string& values)
{
  lattice_vectors lv;
  double ca,cb,ra,rb,phi;
  std::vector<double> lp;
  char_separator<char> sep(" ,"); 
  tokenizer<char_separator<char> > lp_tokens(values,sep);
  tokenizer<char_separator<char> >::iterator it; 
  for (it=lp_tokens.begin();it!=lp_tokens.end();it++)
    lp.push_back(atof((*it).c_str()));
  if (lp.size()<6) throw po::invalid_option_value("This entry must contain all \
six values a,b,c,alpha,beta and gamma, separated by spaces or comma's.");
  
  lp[ALPHA]*=rad; lp[GAMMA]*=rad; lp[BETA]*=rad;
  //std::cout<<lp[ALPHA]<<lp[BETA]<<lp[GAMMA]<<std::endl;
  lv[B][0] = cos(lp[GAMMA]);
  lv[B][1] = sin(lp[GAMMA]);
  lv[B][2] = 0;
  lv[A] = xu;
  ca = lp[C]*cos(lp[BETA]); 
  cb = lp[C]*cos(lp[ALPHA]);
  ra = lp[C]*sin(lp[BETA]); 
  rb = lp[C]*sin(lp[ALPHA]);
  phi = acos( (ca-cb*cos(lp[GAMMA])) / (rb*sin(lp[GAMMA])) );	     
  lv[C][0] = ca;
  lv[C][1] = cb*sin(lp[GAMMA])-rb*cos(lp[GAMMA])*cos(phi);
  lv[C][2] = rb*sin(phi);
  lv[A]*= lp[A];
  lv[B]*= lp[B];
  //std::cout<<lv[A]<<lv[B]<<lv[C]<<std::endl;
  lv[A] = fixed(lv[A],7);lv[B]=fixed(lv[B],7);lv[C]=fixed(lv[C],7);
  //std::cout<<lv[A]<<lv[B]<<lv[C]<<std::endl;
  return lv;
}

lattice_vectors get_reciprocal(lattice_vectors lv)
{
  lattice_vectors lvr;
  lvr[A]=2*M_PI*outer_product(lv[B],lv[C])/
    ublas::inner_prod(lv[A],outer_product(lv[B],lv[C]));
  lvr[B]=2*M_PI*outer_product(lv[C],lv[A])/
    ublas::inner_prod(lv[B],outer_product(lv[C],lv[A]));
  lvr[C]=2*M_PI*outer_product(lv[A],lv[B])/
    ublas::inner_prod(lv[C],outer_product(lv[A],lv[B]));
  return lvr;
}
