/* 
file: lattice_graph.hpp 
Type defenition of the lattice_graph structure. This datastructure 
contains the lattice of spins.

Xpinner v.2, a Monte Carlo program for the simulation of 'classical'
magnetic systems.

Written by Mark de Vries, markadevries@gmail.com

Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of
Edinburgh, Old College, Edinburgh, UK  

Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the 
Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef LATTICE_GRAPH
#define LATTICE_GRAPH

#include <boost/graph/graph_traits.hpp>  // for boost::graph_traits
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/graph/graph_utility.hpp>
#include <boost/program_options.hpp>
#include <iostream>
#include <boost/tokenizer.hpp>
#include <boost/format.hpp>
#include <utility>

#include "math_utility.hpp"

using namespace boost;
using namespace boost::numeric;
namespace po = program_options;

//a lattice site with a magnetic atom/ion
struct vertex_data {
  double mu_eff;
  euclid_vector spin;
  euclid_vector x;
  euclid_vector cell_index;
  euclid_vector fract_coords;
  int site_index;
  std::string label;
};

//a connection between to atoms, can be a magnetic bond if J!=0
//otherwise interaction is only via the dipole-dipole coupling.
struct edge_data {
  double J,r_abs;
  int d;
  euclid_vector ru; //from now on we use a unit vector here. 
  euclid_vector D; //this one is for DMI.
  euclid_vector cell_offset;
};

//graph definitions
//first template param is edge_list type, second vertex_list type.
typedef adjacency_list<setS, vecS, undirectedS, vertex_data
		       ,edge_data > lattice_graph;
typedef graph_traits<lattice_graph>::vertex_iterator site_iterator;
typedef graph_traits<lattice_graph>::edge_iterator edge_iterator;
typedef graph_traits<lattice_graph>::out_edge_iterator o_edge_iterator;
typedef graph_traits<lattice_graph>::vertex_descriptor Site;
typedef graph_traits<lattice_graph>::edge_descriptor Edge;

typedef ublas::bounded_vector<euclid_vector,3> lattice_vectors;

template<class Tf,class Ts>
std::ostream& operator<<(std::ostream& ostr,std::pair<Tf,Ts>& p)
{
  ostr<<" ("<<p.first<<", "<<p.second<<") ";
  return ostr;
} 

template<class Tf,class Ts>
std::pair<Tf,Ts>& operator>>(std::istream& istr,std::pair<Tf,Ts>& p)
{
  std::string s1,s2,s3;
  istr>>s1>>p.first>>s2<<p.second>>s3;
  return p;
} 

std::ostream& operator<<(std::ostream& ostr,lattice_graph& lg);

lattice_graph& operator>>(std::istream& istr,lattice_graph& lg);

std::string out_VECT(lattice_graph& lg);

lattice_vectors get_lattice_vectors(const std::string& values);
lattice_vectors get_reciprocal(lattice_vectors lv);

#endif
 
