/* 
file : math_utility.hpp 
Several generic data type defenitions and math functions.

Spinner v.2, a Monte Carlo program for the simulation of 'classical'
magnetic systems.

Written by Mark de Vries, markadevries@gmail.com

Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of
Edinburgh, Old College, Edinburgh, UK  

Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the 
Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef MATH_UTILS
#define MATH_UTILS

#define NDEBUG

#include <stdlib.h>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <map>
#include <vector>
#include <deque>
#include <math.h>
#include <complex>

using namespace boost;
using namespace boost::numeric;

enum labels_abc {A,B,C,ALPHA,BETA,GAMMA};
enum labels_xyz {X,Y,Z};

const ublas::unit_vector<double> xu(3,X);
const ublas::unit_vector<double> yu(3,Y);
const ublas::unit_vector<double> zu(3,Z);
const ublas::scalar_vector<double> unity(3,1);
const ublas::identity_matrix<double> identity(3);
const ublas::zero_matrix<double> M_zero(3,3);
const ublas::zero_vector<double> zero(3);
const double rad=M_PI/180;

typedef ublas::bounded_vector<double,3> euclid_vector;
typedef ublas::c_matrix<double,3,3> euclid_matrix;
typedef ublas::bounded_vector<uint,3> index_vector;
const std::deque<double> empty_fifo;
const std::vector<euclid_vector> empty_vector_of_euclid_vector;
typedef unsigned int uint;
typedef std::map<std::string,std::deque<double> > map_of_fifos; 

//inline euclid_vector fabs(const euclid_vector& v);
inline double sign(double a);
inline double fixed(double a,int decimals);
inline euclid_vector fixed(const euclid_vector& a,int decimals);
inline euclid_vector floor(const euclid_vector& a);
inline euclid_vector operator%(const euclid_vector& a,const euclid_vector& b);
//inline euclid_vector operator*(const euclid_vector& a, const euclid_vector& b);
inline bool operator==(const euclid_vector& a, const euclid_vector& b);

struct shortest_bond : public std::binary_function<euclid_vector,euclid_vector,bool>
{
  bool operator()(euclid_vector a, euclid_vector b) { 
    return ublas::norm_2(a)< ublas::norm_2(b);
  } 
};

//euclid_vector outer_product(euclid_vector& a,euclid_vector& b);

template<class T>
std::ostream& operator<<(std::ostream& os, const std::vector<T>& v)
{
  for (int n=0; n<v.size(); n++){
    os<<v[n]<<'\t';
    if ((n+1)%8==0) os<<'\n';
  }
  return os;
}

template<class T>
std::ostream& operator<<(std::ostream& os, const std::deque<T>& v)
{
  for (int n=0; n<v.size(); n++)
    os<<v[n]<<'\t';
  return os;
}

inline void clear(map_of_fifos::iterator b,map_of_fifos::iterator e)
{
  //Clear memory of old data.
  for (;b!=e; b++) b->second.erase(b->second.begin(),b->second.end());
}

inline euclid_vector outer_product(const euclid_vector& a,const euclid_vector& b)
{
  euclid_vector c;
  if ((a.size() < 3) or (b.size() < 3))
    std::cerr<<"Warning: implementation error. Outer product used on vectors with dim < 3!"<<std::endl;
  c[0]=a[1]*b[2]-a[2]*b[1];
  c[1]=a[2]*b[0]-a[0]*b[2];
  c[2]=a[0]*b[1]-a[1]*b[0];
  return c;
}

inline double sqr(const euclid_vector& a)
{
  return ublas::inner_prod(a,a);
}

/*
inline euclid_vector fabs(const euclid_vector& v)
{ // What is this one doing??? That cannot be right? Use pythagoras!!
  euclid_vector w=zero;
  w[0]=fabs(v[0]); w[1]=fabs(v.2]); w[2]=fabs(v[2]);
  return w;
} */

inline double sign(double a)
{
  if (a >= 0) return 1;
  else return -1;
}

inline double fixed(double a,int decimals)
{
  double E=pow(10,decimals);
  return a=1.0*round(E*a)/E;
}

inline euclid_vector fixed(const euclid_vector& a,int decimals)
{
  euclid_vector b;
  for (int i=0; i<3; i++) b[i]=fixed(a[i],decimals);
  return b;
}

inline euclid_vector floor(const euclid_vector& a)
{
  euclid_vector b;
  for (int i=0; i<3; i++) b[i]=round(a[i]);
  return b;
}


inline bool operator==(const euclid_vector& a,const euclid_vector& b)
{
  return (a[0]==b[0]&&a[1]==b[1]&&a[2]==b[2]);
}


inline euclid_vector operator%(const euclid_vector& a,const euclid_vector& b)
{
  euclid_vector m;
  m[0]=int(a[0])%int(b[0]); 
  m[1]=int(a[1])%int(b[1]); 
  m[2]=int(a[2])%int(b[2]);
  return m;
}

inline euclid_matrix RzRxRz(double alpha, double beta, double gamma)
{
  euclid_matrix Rzxz,Rz1=M_zero,Rz2=M_zero;
  euclid_matrix Rx=M_zero;
  Rz1(0,0) = cos(alpha); Rz1(0,1) = -sin(alpha);
  Rz1(1,0) = sin(alpha); Rz1(1,1) = cos(alpha);

  Rz2(0,0) = cos(gamma); Rz2(0,1) = -sin(gamma);
  Rz2(1,0) = sin(gamma); Rz2(1,1) = cos(gamma);

  Rx(1,1) = cos(beta); Rx(1,2) = sin(beta);
  Rx(2,1) =-sin(beta); Rx(2,2) = cos(beta);
  Rzxz = ublas::prod(Rx,Rz1);
  Rzxz = ublas::prod(Rz2,Rzxz);
  return Rzxz;
}

inline euclid_vector Rz(double th, const euclid_vector& a)
{
  euclid_vector b;
  b[0] = cos(th)*a[0] - sin(th)*a[1];
  b[1] = sin(th)*a[0] + cos(th)*a[1];
  b[2] = a[2];
  return b;
}

inline euclid_vector Ry(double th, const euclid_vector& a)
{
  euclid_vector b;
  b[0] = cos(th)*a[0] + sin(th)*a[2];
  b[1] = a[1];
  b[2] = -sin(th)*a[0] + cos(th)*a[2];
  return b;
}

#endif
